︠1c9525b0-35f7-4f99-9370-f833253cd937︠
r"""

q_lib - library of numerical quaternion functions

Authors:
    -- Doug Sweetser (2009): Initial version.
    
Examples:
    Parts
    sage: q1 = q(1, 2, 3, 4)
    sage: q1.scalar() == q(1, 0, 0, 0)
    True
    sage: q1.vector() == q(0, 2, 3, 4)
    True
    sage: q1.norm() == q(30, 0, 0, 0)
    True
     
    Add quaternions 3 times
    sage: q1 = q(1,1,1,1); q2 = q(1,2,3,4); q3 = q(3, 5, 7, 9)
    sage: test = add_n(q2, 3)
    sage: q3 == test
    True
    
    Test the various conjugates.
    sage: q1 = q(1, 2, 3, 4); q_time_rev = q(-1, 2, 3, 4); 
    sage: q_space_rev = q(1, -2, -3, -4); q_spacetime_rev = q(-1, -2, -3, -4)
    sage: q1.conj() == q_space_rev
    True
    sage: q1.conj().negate() == q_time_rev
    True
    sage: q1.negate() == q_spacetime_rev
    True
    
    Groups
    sage: q1 = q(2, 1, 2, 0); q2 = q(2/3, 1/3, 2/3, 0)
    sage: q1.u1() == q2
    True
    
    Test the group U(1)xSU(2).
    sage: q1 = q(0, 2, 1, 2); q2 = q(-sin(3), 2 * cos(3)/3, cos(3)/3, 2 * cos(3)/3)
    sage: q1.u1xsu2() == q2
    True
"""
#*****************************************************************************
#       Copyright (C) 2009 Doug Sweetser <sweetser@alum.mit.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#*****************************************************************************
︡95d303cd-6cbe-4482-ba13-b7d145a3a81e︡︡
︠dc27a9a6-1e53-4d93-8f7f-757d332cf02e︠
class q(list):
    def __init__(self,t,x,y,z):
        self.t = t
        self.x = x
        self.y = y
        self.z = z
        self.qn = []
        list.__init__(self,[t,x,y,z])
    def __repr__(self):
        return "Quaternion tools."
    def show(self):
        print self.t, self.x, self.y, self.z
    def normalize(self):
        abs_value = sqrt(self.t^2 + self.x^2 + self.y^2 + self.z^2)
        return q(self.t/abs_value, self.x/abs_value, self.y/abs_value, self.z/abs_value)
        
    # Parts and sizes    
    def scalar(self):
        return q(self.t, 0, 0, 0)
    def vector(self):
        return q(0, self.x, self.y, self.z)
    def norm(self):
        return q(self.t * self.t + self.x * self.x + self.y * self.y + self.z * self.z, 0, 0, 0)

    # Conjugates
    def conj(self):
        return q(self.t, -self.x, -self.y, -self.z)
    def conj_1(self):
        return q(-self.t, self.x, -self.y, -self.z)   
    def conj_2(self):
        return q(-self.t, -self.x, self.y, -self.z)
    def conj_3(self):
        return q(-self.t, -self.x, -self.y, self.z)   
    def negate(self):
        return q(-self.t, -self.x, -self.y, -self.z) 

    def add(self, q1):
        return q(self.t + q1.t, self.x + q1.x, self.y + q1.y, self.z + q1.z)
    def add_n(self, q1, n):
        self.qn[:]=[]
        for i in range (0, n):
            self.t += q1.t
            self.x += q1.x
            self.y += q1.y
            self.z += q1.z
            self.qn.append(q(self.t, self.x, self.y, self.z))
        return q(self.t, self.x, self.y, self.z)
    def scalar_div(self, d):
        return(q(self.t/d, self.x/d, self.y/d, self.z/d));
    def scalar_x(self, d):
        return(q(self.t * d, self.x * d, self.y * d, self.z * d));

    def real(self):
        t = self.add(self.conj()).scalar_div(2)
        return q(t.t, t.x, t.y, t.z)
    def complex_x(self):
         tx =self.add(self).add(self.conj()).add(self.conj_1()).scalar_div(2)
         return q(tx.t, tx.x, tx.y, tx.z)
    def complex_y(self):
         ty = self.add(self).add(self.conj()).add(self.conj_2()).scalar_div(2)
         return q(ty.t, ty.x, ty.y, ty.z)
    def complex_z(self):
         tz = self.add(self).add(self.conj()).add(self.conj_3()).scalar_div(2)
         return q(tz.t, tz.x, tz.y, tz.z)
            
    def n_random(self, n):
        self.qn[:]=[]
        for i in range(n):
            self.qn.append(q(2 * random() - 1, 2 * random() - 1, 2 * random() - 1, 2 * random() - 1))
        return self
        
    def times(self, q2):
        return q(self.t * q2.t - self.x * q2.x - self.y * q2.y - self.z * q2.z, self.t * q2.x + self.x * q2.t + self.y * q2.z - self.z * q2.y, self.t * q2.y + self.y * q2.t + self.z * q2.x - self.x * q2.z, self.t * q2.z + self.z * q2.t + self.x * q2.y - self.y * q2.x)
        
# Exponentials and trig functions
#   exp(q) = (exp(t) cos(|R|), exp(t) sin(|R|) R/|R|)
    def exp(self):
        abs_vector = sqrt(self.x^2 + self.y^2 + self.z^2)
        if abs_vector == 0:
            return q(exp(self.t), 0, 0, 0)
        k = exp(self.t) * sin(abs_vector) / abs_vector
        return q(exp(self.t) * cos(abs_vector), k * self.x, k * self.y, k * self.z)

#   sin(q) = (sin(t) cosh(|R|), cos(t) sinh(|R|) R/|R|)
    def sin(self):
        abs_vector = sqrt(self.x^2 + self.y^2 + self.z^2)
        if abs_vector == 0:
            return q(sin(self.t), 0, 0, 0)
        k = cos(self.t) * sinh(abs_vector) / abs_vector
        return q(sin(self.t) * cosh(abs_vector), k * self.x, k * self.y, k * self.z)
        
#   cos(q) = (cos(t) cosh(|R|), sin(t) sinh(|R|) R/|R|)
    def cos(self):
        abs_vector = sqrt(self.x^2 + self.y^2 + self.z^2)
        if abs_vector == 0:
            return q(cos(self.t), 0, 0, 0)
        k = sin(self.t) * sinh(abs_vector) / abs_vector
        return q(cos(self.t) * cosh(abs_vector), k * self.x, k * self.y, k * self.z)
        
# Groups

    def u1(self):
        return self.normalize()
    def su2(self):
        return self.vector().exp()
    def u1xsu2(self):
        return self.su2().times(self.u1())

#auto
︡69416970-7b4c-4207-85f5-12fc6ebf6c97︡︡
︠4598a1cc-00b9-467f-be96-46e7fbb9484d︠
r"""Check all testing code"""
︡6241a400-90b3-4203-bfe7-f1d1fec429dc︡{"html": "<span class=\"math\">\\text{Check all testing code}</span>"}︡
︠892c667e-9eec-40bf-8fd7-fdc4f7b853d8︠
sage: q1 = q(1, 2, 3, 4)
sage: q1.scalar() == q(1, 0, 0, 0)
︡7c0ca0dc-d333-45bb-bc47-f33cf43da414︡{"html": "<span class=\"math\">{\\rm True}</span>"}︡
︠d86be997-f9ba-4776-bd40-6f9b63d945e1︠
sage: q1.vector() == q(0, 2, 3, 4)
︡6947dd7c-dbf1-4325-9803-b337a6f66c5d︡{"html": "<span class=\"math\">{\\rm True}</span>"}︡
︠f51611c8-8d20-42c6-aece-7310d35a2d78︠
sage: q1.norm() == q(30, 0, 0, 0)
︡c38b7ea5-016b-4ac6-a633-2f7821bbb406︡{"html": "<span class=\"math\">{\\rm True}</span>"}︡
︠b7e1652b-febf-4a06-8c98-57d30b75e37a︠
sage: q1.real()==q(1, 0, 0, 0)
︡9ac1e0c8-8278-4344-ac1d-09968fdcf2b7︡{"html": "<span class=\"math\">{\\rm True}</span>"}︡
︠9cc75c30-d2eb-43f6-ac35-2de3544900c4︠
sage: q1.complex_x()==q(1, 2, 0, 0)
︡d29b6f85-b24d-4752-9eae-46ce2ca1ec3a︡{"html": "<span class=\"math\">{\\rm True}</span>"}︡
︠f94ce526-ec1b-4a58-9955-08893e7cfbd4︠
sage: q1.complex_y()==q(1, 0, 3, 0)
︡66bb1abd-453b-49f9-b140-d4e5bd2f5949︡{"html": "<span class=\"math\">{\\rm True}</span>"}︡
︠aae96e59-27a6-4ebf-9870-7df1d814ed6e︠
sage: q1.complex_z()==q(1, 0, 0, 4)
︡1336040a-10bf-48e5-ad77-e3a5c65705a4︡{"html": "<span class=\"math\">{\\rm True}</span>"}︡
︠c545df08-f7c8-47e7-912a-0cfacd66c6b6︠
sage: q1 = q(1, 2, 3, 4); q_time_rev = q(-1, 2, 3, 4); 
sage: q_space_rev = q(1, -2, -3, -4); q_spacetime_rev = q(-1, -2, -3, -4)
sage: q1.conj() == q_space_rev
︡3195f66c-125f-48c7-80d1-c3efb0b3a385︡{"html": "<span class=\"math\">{\\rm True}</span>"}︡
︠c56277d9-ffc3-4626-99a0-5c9d8369c172︠
sage: q1.conj().negate() == q_time_rev
︡0b216240-2cd9-445d-b13c-4c2014150a2c︡{"html": "<span class=\"math\">{\\rm True}</span>"}︡
︠643f8e65-544d-44d8-99f6-d68438e55297︠
sage: q1.negate() == q_spacetime_rev
︡f8716656-3f7b-4127-b6c9-6b406f37b06c︡{"html": "<span class=\"math\">{\\rm True}</span>"}︡
︠ef98e84b-09d7-48ce-a06c-148c26b3e8be︠
q1 = q(1,1,1,1); q2 = q(1,2,3,4); q3 = q(3, 5, 7, 9)
test = q1.add_n(q2, 2)
test == q3
︡1b47904c-7bd3-4311-bb10-6cbfc93f46d1︡{"html": "<span class=\"math\">{\\rm True}</span>"}︡
︠87299fef-e0f6-4e75-a4b7-90e7e8b79c68︠
sage: q1 = q(2, 1, 2, 0); q2 = q(-79/81, -8/81, -16/81, 0)
sage: q1.group_u1(2) == q2
︡34434e90-ac59-48c6-bf0e-15973d91334d︡{"html": "<span class=\"math\">{\\rm True}</span>"}︡
︠8a2907d8-7cd8-42aa-ba86-591cf2fcda5a︠
q1 = q(2, 1, 2, 0).u1().show()
︡969b7f1c-a3fe-47b0-82dd-5e473ca3a538︡{"stdout": "2/3 1/3 2/3 0"}︡
